import hashlib
import json
import lzma
import pickle
import urllib
import urllib.error
import urllib.parse
import urllib.request
import urllib.response
from abc import ABC
from concurrent.futures import Future, ThreadPoolExecutor, as_completed
from http.client import HTTPResponse
from pathlib import Path
from typing import List, Literal, Optional, Type, TypeVar, Union, overload

from rich.progress import BarColumn, Progress, TimeRemainingColumn

from fast_bioservices._log import logger
from fast_bioservices._settings import cache_dir

Method = Literal["GET"]
ResponseType = TypeVar("ResponseType", bound="Response")

class Response:
    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, "_from_class_method"):
            raise NotImplementedError("Use `from_response` or `from_cache` to create a Response object")
        return super().__new__(cls)

    def __init__(self) -> None:
        self._url: str
        self._headers: dict
        self._debug_level: int
        self._content: bytes
        self._status: int
        self._method: Method

    @property
    def url(self) -> str:
        return self._url

    @property
    def headers(self) -> dict:
        return self._headers

    @property
    def bytes(self) -> bytes:
        return self._content

    @property
    def text(self) -> str:
        return self.bytes.decode()

    @property
    def json(self) -> dict:
        return json.loads(self.text)

    def cache_object(self) -> dict:
        return {
            "url": self._url,
            "headers": self._headers,
            "debug_level": self._debug_level,
            "content": self._content,
            "status": self._status,
            "method": self._method,
        }

    @overload
    @classmethod
    def create(cls: Type[ResponseType], response: HTTPResponse, method: Method) -> "Response": ...

    @overload
    @classmethod
    def create(cls: Type[ResponseType], *, data: dict) -> "Response": ...  # Use `*` to force keyword-only arguments

    @classmethod
    def create(cls: Type[ResponseType], response=None, method=None, data=None) -> "Response":
        if response is not None and method is not None:
            return cls._from_response(response, method)
        elif data is not None:
            return cls._from_cache(data)
        else:
            raise ValueError("Either (`response` and `method`) or (`data`) must be provided")

    @classmethod
    def _from_response(cls, response: HTTPResponse, method: Method) -> "Response":
        # Ensure that the classmethod is called from `create`
        cls._from_class_method = True
        instance = cls()
        delattr(cls, "_from_class_method")

        instance._url = response.geturl()
        instance._headers = dict(response.getheaders())
        instance._debug_level = response.debuglevel
        instance._content = response.read()
        instance._status = response.getcode()
        instance._method = method
        return instance

    @classmethod
    def _from_cache(cls, data: dict) -> "Response":
        # Ensure that the classmethod is called from `create`
        cls._from_class_method = True
        instance = cls()
        delattr(cls, "_from_class_method")

        instance._url = data["url"]
        instance._headers = data["headers"]
        instance._debug_level = data["debug_level"]
        instance._content = data["content"]
        instance._method = data["method"]
        return instance

class FastHTTP(ABC):
    def __init__(
        self,
        cache: bool,
        workers: int,
        show_progress: bool,
        max_requests_per_second: Optional[int],
    ) -> None:
        self._max_requests_per_second: int = int(1e10) if max_requests_per_second is None else max_requests_per_second
        self._maximum_allowed_workers: int = 16
        self._use_cache: bool = cache
        self._show_progress: bool = show_progress
        self._workers: int = self._set_workers(workers)

        self._cache_dirpath: Path = cache_dir
        self._requests_made: int = 0
        self._last_request_time: float = 0

        self._thread_pool = ThreadPoolExecutor(max_workers=self._workers)
        self._progress = Progress(
            "[progress.description]{task.description}",
            BarColumn(),
            "{task.completed}/{task.total} batches",
            "[progress.percentage]{task.percentage:>3.0f}%",
            TimeRemainingColumn(),
        )

    def _set_workers(self, value: int) -> int:
        if value < 1:
            logger.debug("`max_workers` must be greater than 0, setting to 1")
            value = 1
        elif value > self._maximum_allowed_workers:
            logger.debug(
                f"`max_workers` must be less than {self._maximum_allowed_workers} (received {value}), setting to {self._maximum_allowed_workers}"
            )
            value = self._maximum_allowed_workers
        return value

    @property
    def workers(self) -> int:
        return self._workers

    @workers.setter
    def workers(self, value: int) -> None:
        self._workers = self._set_workers(value)

    def __del__(self):
        self._thread_pool.shutdown(wait=True)

    @property
    def show_progress(self) -> bool:
        return self._show_progress

    @show_progress.setter
    def show_progress(self, value: bool) -> None:
        self._show_progress = value

    def _make_safe_url(self, urls: Union[str, List[str]]) -> List[str]:
        # Safe characters from https://stackoverflow.com/questions/695438
        safe = "&$+,/:;=?@#"
        if isinstance(urls, str):
            return [urllib.parse.quote(urls, safe=safe)]
        return [urllib.parse.quote(url, safe=safe) for url in urls]

    def _calculate_cache_key(self, url: str) -> Path:
        cache_key = hashlib.md5(url.encode()).hexdigest()
        cache_file = Path(self._cache_dirpath, cache_key)
        return cache_file

    def __get_without_cache(self, url: str, headers: dict) -> Response:
        request = urllib.request.Request(url, headers=headers)
        response: HTTPResponse = urllib.request.urlopen(request)
        return Response.create(response=response, method="GET")

    def __get_from_cache(self, url: str, headers: dict) -> Optional[Response]:
        cache_file = self._calculate_cache_key(url)
        if cache_file.exists():
            logger.debug(f"Cache hit: {url}")
            with lzma.open(cache_file, "rb") as cache_file:
                return Response.create(data=pickle.load(cache_file))
        return None

    def _save_to_cache(self, response: Response) -> None:
        cache_filepath = self._calculate_cache_key(response.url)
        with lzma.open(cache_filepath, "wb") as cache_file:
            pickle.dump(response.cache_object(), cache_file)

    def _get(
        self,
        urls: Union[str, List[str]],
        headers: Optional[dict] = None,
        temp_disable_cache: bool = False,
    ) -> List[Response]:
        headers = headers or {}
        urls = self._make_safe_url(urls)
        task = self._progress.add_task("[cyan]Working...", total=len(urls)) if self._progress else None
        callable = self.__get_from_cache if self._use_cache and not temp_disable_cache else self.__get_without_cache

        responses: List[Response] = []
        futures: List[Future[Union[Response, None]]] = [
            self._thread_pool.submit(callable, url, headers) for (url, headers) in zip(urls, [headers] * len(urls))
        ]

        for url, future in zip(urls, as_completed(futures)):
            response = future.result()
            if response is not None:
                responses.append(response)
            else:
                responses.append(self.__get_without_cache(url, headers))
                if self._use_cache:
                    self._save_to_cache(responses[-1])
            if self._progress and task is not None:
                self._progress.update(task, advance=1)

        return responses


if __name__ == "__main__":
    http = FastHTTP(cache=True, workers=2, show_progress=True, max_requests_per_second=10)



from unittest.mock import patch

import pytest
from fast_bioservices.ensembl.comparative_genomics import GetHomology, HomologyResult


@pytest.fixture(scope="session")
def get_homology_instance():
    return GetHomology(cache=False, max_workers=1, show_progress=False)

def test_get_homology_instance_creation(get_homology_instance):
    assert get_homology_instance is not None
    assert get_homology_instance._max_workers == 1
    assert not get_homology_instance._show_progress


def test_get_homology_url_property(get_homology_instance):
    assert get_homology_instance.url == "https://rest.ensembl.org"


@patch("fast_bioservices.ensembl.comparative_genomics.GetHomology._get")
def test_by_species_with_symbol_or_id_returns_homology_results(mock_get, get_homology_instance):
    mock_response = {
        "data": [
            {
                "id": "ENSG00000157764",
                "homologies": [
                    {
                        "method_link_type": "ENSEMBL_ORTHOLOGUES",
                        "dn_ds": None,
                        "taxonomy_level": "Catarrhini",
                        "type": "ortholog_one2one",
                        "target": {
                            "align_seq": "MAALSGGGGGGAEPGQALFNGDMEPE----AGAAASSAADPAIPEEVWNIKQMIKLTQEHIEALLDKFGGEHNPPSIYLEAYEEYTSKLDALQQREQQLLEYLGNGTDFSVSSSASMDTVTSSSSSSLSVLPSSLSVFQNPTDVSRSNPKSPQKPIVRVFLPNKQRTVVPARCGVTVRDSLKKALMMRGLIPECCAVYRIQDGEKKPIGWDTDISWLTGEELHVEVLENVPLTTHNFVRKTFFTLAFCDFCRKLLFQGFRCQTCGYKFHQRCSTEVPLMCVNYDQLDLLFVSKFFEHHPIPQEEASLAETALTSGSSPSAPTSDSLGPQILTSPSPSKSIPIPQPFRPADEDHRNQFGQRDRSSSAPNVHINTIEPVNIDDLIRDQGFRGDGGSTTGLSATPPASLPGSLTNVKALQKSPGPQRERKSSSSSEDRNRMKTLGRRDSSDDWEIPDGQITVGQRIGSGSFGTVYKGKWHGDVAVKMLNVTAPTPQQLQAFKNEVGVLRKTRHVNILLFMGYSTKPQLAIVTQWCEGSSLYHHLHIIETKFEMIKLIDIARQTAQGMDYLHAKSIIHRDLKSNNIFLHEDLTVKIGDFGLATVKSRWSGSHQFEQLSGSILWMAPEVIRMQDKNPYSFQSDVYAFGIVLYELMTGQLPYSNINNRDQIIFMVGRGYLSPDLSKVRSNCPKAMKRLMAECLKKKRDERPLFPQILASIELLARSLPKIHRSASEPSLNRAGFQTEDFSLYACASPKTPIQAGGYGEFAAFK",
                            "id": "ENSMMUG00000042793",
                            "species": "macaca_mulatta",
                            "perc_id": 98.8204,
                            "perc_pos": 99.0826,
                            "cigar_line": "26M4D737M",
                            "taxon_id": 9544,
                            "protein_id": "ENSMMUP00000044832",
                        },
                        "source": {
                            "perc_pos": 98.6945,
                            "cigar_line": "766MD",
                            "taxon_id": 9606,
                            "protein_id": "ENSP00000493543",
                            "align_seq": "MAALSGGGGGGAEPGQALFNGDMEPEAGAGAGAAASSAADPAIPEEVWNIKQMIKLTQEHIEALLDKFGGEHNPPSIYLEAYEEYTSKLDALQQREQQLLESLGNGTDFSVSSSASMDTVTSSSSSSLSVLPSSLSVFQNPTDVARSNPKSPQKPIVRVFLPNKQRTVVPARCGVTVRDSLKKALMMRGLIPECCAVYRIQDGEKKPIGWDTDISWLTGEELHVEVLENVPLTTHNFVRKTFFTLAFCDFCRKLLFQGFRCQTCGYKFHQRCSTEVPLMCVNYDQLDLLFVSKFFEHHPIPQEEASLAETALTSGSSPSAPASDSIGPQILTSPSPSKSIPIPQPFRPADEDHRNQFGQRDRSSSAPNVHINTIEPVNIDDLIRDQGFRGDGGSTTGLSATPPASLPGSLTNVKALQKSPGPQRERKSSSSSEDRNRMKTLGRRDSSDDWEIPDGQITVGQRIGSGSFGTVYKGKWHGDVAVKMLNVTAPTPQQLQAFKNEVGVLRKTRHVNILLFMGYSTKPQLAIVTQWCEGSSLYHHLHIIETKFEMIKLIDIARQTAQGMDYLHAKSIIHRDLKSNNIFLHEDLTVKIGDFGLATVKSRWSGSHQFEQLSGSILWMAPEVIRMQDKNPYSFQSDVYAFGIVLYELMTGQLPYSNINNRDQIIFMVGRGYLSPDLSKVRSNCPKAMKRLMAECLKKKRDERPLFPQILASIELLARSLPKIHRSASEPSLNRAGFQTEDFSLYACASPKTPIQAGGYGAFPVH-",
                            "id": "ENSG00000157764",
                            "species": "homo_sapiens",
                            "perc_id": 98.4334,
                        },
                    }
                ],
            }
        ]
    }
    mock_get.return_value = [type("obj", (object,), {"json": mock_response})]
    results = get_homology_instance.by_species_with_symbol_or_id(
        reference_species="human",
        ensembl_id_or_symbol="ENSG00000157764",
        target_species="macaca_mulatta",
    )
    assert isinstance(results, list)
    assert len(results) > 0
    assert isinstance(results[0], HomologyResult)
    assert results[0].id == "ENSG00000157764"


@patch("fast_bioservices.ensembl.comparative_genomics.GetHomology._get")
def test_by_species_with_symbol_or_id_url_construction(mock_get, get_homology_instance):
    get_homology_instance.by_species_with_symbol_or_id(
        reference_species="human",
        ensembl_id_or_symbol="ENSG00000157764",
        aligned=False,
        cigar_line=False,
        compara="vertebrates",
        external_db="test_db",
        format="full",
        sequence="protein",
        target_species="macaca_mulatta",
        target_taxon=123,
        type="orthologues",
    )
    expected_url = [
        "https://rest.ensembl.org/homology/id/human/ENSG00000157764?compara=vertebrates;format=full;sequence=protein;type=orthologues;external_db=test_db;target_species=macaca_mulatta;target_taxon=123"
    ]
    mock_get.assert_called_with(urls=expected_url, headers={"Content-Type": "application/json"})



from difflib import SequenceMatcher
from typing import List, Union


def flatten(lists: list) -> list:
    data = []
    for item in lists:
        if isinstance(item, list):
            data.extend(flatten(item))
        else:
            data.append(item)
    return data


def fuzzy_search(query: str, possibilities: Union[str, List[str]]) -> float:
    matcher = SequenceMatcher(isjunk=None, a=query)
    if isinstance(possibilities, str):
        possibilities = [possibilities]

    # Provide a default value for the ratio
    ratios: list[float] = [0.0]
    for possible in possibilities:
        matcher.set_seq2(possible)
        ratios.append(matcher.ratio())
    return max(ratios)



import logging

from fast_bioservices import _settings

logger = logging.getLogger("fast_bioservices")
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
    filename=_settings.log_filepath,
)



from dataclasses import dataclass
from typing import List, Optional

from fast_bioservices._base import BaseModel
from fast_bioservices._fast_http import FastHTTP
from fast_bioservices._utils import fuzzy_search


@dataclass
class Species:
    strain_collection: str
    aliases: list[str]
    name: str
    taxon_id: int
    assembly: str
    common_name: str
    groups: list[str]
    accession: str
    release: int
    division: str
    strain: str
    display_name: str


@dataclass
class FuzzyResult:
    species: Species
    score: float


class Ensembl(BaseModel, FastHTTP):
    def __init__(
        self,
        max_workers: int,
        show_progress: bool,
        cache: bool = True,
    ):
        self._url = "https://rest.ensembl.org"
        BaseModel.__init__(self, url=self._url)
        FastHTTP.__init__(
            self,
            cache=cache,
            workers=max_workers,
            max_requests_per_second=15,
            show_progress=show_progress,
        )

    @property
    def url(self) -> str:
        return self._url

    def __get_species(self) -> List[Species]:
        path = self._url + "/info/species"
        response = self._get(path, headers={"Content-Type": "application/json"})
        species: list[Species] = []
        for item in response[0].json["species"]:
            species.append(Species(**item))
        return species

    def _match_species(self, species: str) -> Optional[Species]:
        """
        This function will make sure the user enters a valid species name
        It will do this by collecting all the species names from the Ensembl API
        Then doing a fuzzy search on the species name
        If the species name does not exactly match an alias from the API,
        it will show a warning so the user knows they have not perfectly matched a species name

        Parameters
        ----------
        species : str
            The species to validate

        Returns
        -------
        Species
            The species object from the Ensembl API
        """
        species = species.lower()
        species_list = self.__get_species()

        matches: list[FuzzyResult] = []
        for possible_matches in species_list:
            score = fuzzy_search(query=species, possibilities=possible_matches.aliases)
            matches.append(FuzzyResult(possible_matches, score))

        if not matches:
            return None
        highest_match = max(matches, key=lambda x: x.score)
        return highest_match.species


def main():
    e = Ensembl(max_workers=1, show_progress=True)
    e._match_species("human")


if __name__ == "__main__":
    main()


import io
from typing import Dict, List, Literal, Union

import pandas as pd

from fast_bioservices._base import BaseModel
from fast_bioservices._fast_http import FastHTTP, Response
from fast_bioservices._log import logger
from fast_bioservices.biodbnet.nodes import Input, Output, Taxon


class BioDBNet(BaseModel, FastHTTP):
    def __init__(
        self,
        max_workers: int = 4,
        show_progress: bool = True,
        cache: bool = True,
    ):
        self._url = "https://biodbnet-abcc.ncifcrf.gov/webServices/rest.php/biodbnetRestApi.json"
        self._chunk_size: int = 250

        BaseModel.__init__(self, url=self._url)
        FastHTTP.__init__(
            self,
            cache=cache,
            workers=max_workers,
            max_requests_per_second=10,
            show_progress=show_progress,
        )

    @property
    def url(self) -> str:
        return self._url

    def _are_nodes_valid(
        self,
        input_: Union[Input, Output],
        output: Union[Input, Output, List[Input], List[Output]],
        direct_output: bool = False,
    ) -> bool:
        """
        The input database and output database must be different.

        Parameters
        ----------
        input_ : Union[Input, Output]
            The input database
        output : Union[Input, Output, List[Input], List[Output]]
            The output database
        direct_output : bool, optional
            Get direct output node(s) for a given input node (i.e., outputs reacable by a single connection), by default False

        Returns
        -------
        bool
            True if the input and output databases are different, False otherwise.
        """

        logger.debug("Validating databases")
        output_list = [output] if not isinstance(output, list) else output

        if direct_output:
            return all(
                [o.value in self.getDirectOutputsForInput(input_) for o in output_list]
            )
        return all([o.value in self.getOutputsForInput(input_) for o in output_list])

    def _validate_taxon_id(
        self,
        taxon: Union[int, Taxon, List[Union[int, Taxon]]],
    ) -> Union[int, List[int]]:
        taxon_list: list = taxon if isinstance(taxon, list) else [taxon]
        for i in range(len(taxon_list)):
            if isinstance(taxon_list[i], Taxon):
                taxon_list[i] = taxon_list[i].value

            logger.debug(f"Validating taxon ID '{taxon_list[i]}'")
            taxon_url: str = (
                f"https://www.ncbi.nlm.nih.gov/taxonomy/?term={taxon_list[i]}"
            )
            if (
                "No items found."
                in self._get(taxon_url, temp_disable_cache=True)[0].text
            ):
                raise ValueError(f"Unable to find taxon '{taxon_list[i]}'")
        logger.debug(f"Taxon IDs are valid: {','.join([str(i) for i in taxon_list])}")

        return taxon_list[0] if len(taxon_list) == 1 else taxon_list

    def getDirectOutputsForInput(self, input: Union[Input, Output]) -> List[str]:
        url = f"{self.url}?method=getdirectoutputsforinput&input={input.value.replace(' ', '').lower()}&directOutput=1"
        print(url)
        outputs = self._get(url, temp_disable_cache=True)[0].json
        return outputs["output"]

    def getInputs(self) -> List[str]:
        url = f"{self.url}?method=getinputs"
        inputs = self._get(url, temp_disable_cache=True)[0].json
        return inputs["input"]

    def getOutputsForInput(self, input: Union[Input, Output]) -> List[str]:
        url = f"{self.url}?method=getoutputsforinput&input={input.value.replace(' ', '').lower()}"
        valid_outputs = self._get(url, temp_disable_cache=True)[0].json
        return valid_outputs["output"]

    def getAllPathways(
        self,
        taxon: Union[Taxon, int],
        as_dataframe: bool = False,
    ) -> Union[pd.DataFrame, List[Dict[str, str]]]:
        taxon_id = self._validate_taxon_id(taxon)

        url = f"{self.url}?method=getpathways&pathways=1&taxonId={taxon_id}"
        result = self._get(url)[0].json
        if as_dataframe:
            return pd.DataFrame(result)
        return result  # type: ignore

    def getPathwayFromDatabase(
        self,
        pathways: Union[
            Literal["reactome", "biocarta", "ncipid", "kegg"],
            List[Literal["reactome", "biocarta", "ncipid", "kegg"]],
        ],
        taxon: Union[Taxon, int] = Taxon.HOMO_SAPIENS,
        as_dataframe: bool = True,
    ) -> Union[pd.DataFrame, List[Dict[str, str]]]:
        taxon_id = self._validate_taxon_id(taxon)

        if isinstance(pathways, str):
            pathways = [pathways]

        url = f"{self.url}?method=getpathways&pathways={','.join(pathways)}&taxonId={taxon_id}"
        result = self._get(url)[0].json

        if as_dataframe:
            return pd.DataFrame(result)
        return result  # type: ignore

    def db2db(
        self,
        input_values: List[str],
        input_db: Input,
        output_db: Union[Output, List[Output]],
        taxon: Union[Taxon, int] = Taxon.HOMO_SAPIENS,
    ) -> pd.DataFrame:
        taxon_id = self._validate_taxon_id(taxon)
        if not self._are_nodes_valid(input_db, output_db):
            out_db = [output_db] if not isinstance(output_db, list) else output_db
            raise ValueError(
                "You have provided an invalid output database(s).\n"
                "A common result of this problem is including the input database as an output database.\n"
                f"Input database: {input_db.value}\n"
                f"Output database(s): {','.join([o.value for o in out_db])}"
            )
        logger.debug("Databases are valid")

        if isinstance(output_db, Output):
            output_db_value = output_db.value
        else:
            output_db_value = ",".join([o.value for o in output_db])
        logger.debug(f"Got an input database with a value of '{input_db.value}'")
        logger.debug(
            f"Got {len(output_db_value.split(','))} output databases with values of: '{output_db_value}'"
        )

        urls: list[str] = []
        for i in range(0, len(input_values), self._chunk_size):
            urls.append(self.url + "?method=db2db&format=row")
            urls[-1] += f"&input={input_db.value}"
            urls[-1] += f"&outputs={output_db_value}"
            urls[-1] += (
                f"&inputValues={','.join(input_values[i: i + self._chunk_size])}"
            )
            urls[-1] += f"&taxonId={taxon_id}"

        responses: List[Response] = self._get(urls=urls)
        df = pd.DataFrame()
        for response in responses:
            df = pd.concat([df, pd.DataFrame(response.json)])

        df.rename(columns={"InputValue": input_db.value}, inplace=True)
        logger.debug(f"Returning dataframe with {len(df)} rows")
        return df

    def dbWalk(
        self,
        input_values: List[str],
        db_path: List[Union[Input, Output]],
        taxon: Union[Taxon, int] = Taxon.HOMO_SAPIENS,
    ) -> pd.DataFrame:
        taxon_id = self._validate_taxon_id(taxon)

        for i in range(len(db_path) - 1):
            current_db = db_path[i]
            next_db = db_path[i + 1]

            if not self._are_nodes_valid(current_db, next_db, direct_output=True):
                raise ValueError(
                    "You have provided an invalid output database.\n"
                    f"Unable to navigate from '{current_db.value}' to '{next_db.value}'"
                )
        logger.debug("Databases are valid")
        databases: list[str] = [d.value.replace(" ", "").lower() for d in db_path]

        urls: list[str] = []
        for i in range(0, len(input_values), self._chunk_size):
            urls.append(self.url + "?method=dbwalk&format=row")
            urls[-1] += f"&inputValues={','.join(input_values[i:i + self._chunk_size])}"
            urls[-1] += f"&dbPath={'->'.join(databases)}"
            urls[-1] += f"&taxonId={taxon_id}"

        responses: List[Response] = self._get(urls)
        df = pd.DataFrame()
        for response in responses:
            df = pd.concat([df, pd.DataFrame(response.json)])
        df = df.rename(columns={"InputValue": str(db_path[0].value)})

        logger.debug(f"Returning dataframe with {len(df)} rows")
        return df

    def dbReport(
        self,
        input_values: List[str],
        input_db: Union[Input, Output],
        taxon: Union[Taxon, int] = Taxon.HOMO_SAPIENS,
    ):
        taxon_id = self._validate_taxon_id(taxon)

        urls: list[str] = []
        for i in range(0, len(input_values), self._chunk_size):
            urls.append(self.url + "?method=dbreport&format=row")
            urls[-1] += f"&input={input_db.value.replace(' ', '').lower()}"
            urls[-1] += f"inputValues={','.join(input_values[i:i + self._chunk_size])}"
            urls[-1] += f"&taxonId={taxon_id}"

        return NotImplementedError

    def dbFind(
        self,
        input_values: List[str],
        output_db: Union[Output, List[Output]],
        taxon: Union[Taxon, int] = Taxon.HOMO_SAPIENS,
    ) -> pd.DataFrame:
        if isinstance(output_db, Output):
            output_db = [output_db]

        taxon_id = self._validate_taxon_id(taxon)

        urls: list[str] = []
        for out_db in output_db:
            for i in range(0, len(input_values), self._chunk_size):
                urls.append(self.url + "?method=dbfind&format=row")
                urls[-1] += (
                    f"&inputValues={','.join(input_values[i:i + self._chunk_size])}"
                )
                urls[-1] += f"&output={out_db.value}"
                urls[-1] += f"&taxonId={taxon_id}"

        responses: List[Response] = self._get(urls=urls)
        df = pd.DataFrame()
        for response in responses:
            df = pd.concat([df, pd.DataFrame(response.json)])
        return df

    def dbOrtho(
        self,
        input_values: List[str],
        input_db: Input,
        output_db: Union[Output, List[Output]],
        input_taxon: Union[Taxon, int] = Taxon.HOMO_SAPIENS,
        output_taxon: Union[Taxon, int] = Taxon.MUS_MUSCULUS,
    ):
        input_taxon_value = self._validate_taxon_id(input_taxon)
        output_taxon_value = self._validate_taxon_id(output_taxon)

        if isinstance(output_db, Output):
            output_db = [output_db]

        urls: list[str] = []
        for out_db in output_db:
            for i in range(0, len(input_values), self._chunk_size):
                urls.append(self.url + "?method=dbortho")
                urls[-1] += f"&input={input_db.value.replace(' ', '').lower()}"
                urls[-1] += (
                    f"&inputValues={','.join(input_values[i:i + self._chunk_size])}"
                )
                urls[-1] += f"&inputTaxon={input_taxon_value}"
                urls[-1] += f"&outputTaxon={output_taxon_value}"
                urls[-1] += f"&output={out_db.value.replace(' ', '').lower()}"
                urls[-1] += "&format=row"

        responses: List[Response] = self._get(urls=urls)
        df = pd.DataFrame()
        for response in responses:
            df = pd.concat([df, pd.DataFrame(response.json)])

        # Remove potential duplicate columns
        for column in df.columns:
            if str(column).endswith("_x"):
                df = df.drop(column, axis=1)
            elif str(column).endswith("_y"):
                df.rename(columns={column: column[:-2]}, inplace=True)

        df.rename(columns={"InputValue": input_db.value}, inplace=True)

        return df

    def dbAnnot(
        self,
        input_values: List[str],
        annotations: List[
            Literal[
                "Drugs",
                "Diseases",
                "Genes",
                "GO Terms",
                "Pathways",
                "Protein Interactors",
            ]
        ],
        taxon: Union[Taxon, int] = Taxon.HOMO_SAPIENS,
    ) -> pd.DataFrame:
        taxon_id = self._validate_taxon_id(taxon)

        annotations_ = [a.replace(" ", "").lower() for a in annotations]
        urls: list[str] = []
        for i in range(0, len(input_values), self._chunk_size):
            urls.append(self.url + "?method=dbannot")
            urls[-1] += f"&inputValues={','.join(input_values[i:i + self._chunk_size])}"
            urls[-1] += f"&taxonId={taxon_id}"
            urls[-1] += f"&annotations={','.join(annotations_)}"
            urls[-1] += "&format=row"

        responses: list[Response] = self._get(urls=urls)
        df = pd.DataFrame()
        for response in responses:
            df = pd.concat([df, pd.DataFrame(response.json)])
        df = df.rename(columns={"InputValue": "Input Value"})
        return df

    def dbOrg(
        self,
        input_db: Input,
        output_db: Output,
        taxon: Union[Taxon, int] = Taxon.HOMO_SAPIENS,
    ) -> pd.DataFrame:
        taxon_id = self._validate_taxon_id(taxon)

        input_db_val = input_db.value.replace(" ", "_")
        output_db_val = output_db.value.replace(" ", "_")

        url = f"https://biodbnet-abcc.ncifcrf.gov/db/dbOrgDwnld.php?file={input_db_val}__to__{output_db_val}_{taxon_id}"
        buffer = io.StringIO(self._get(url)[0].text)
        return pd.read_csv(
            buffer, sep="\t", header=None, names=[input_db.value, output_db.value]
        )


if __name__ == "__main__":
    biodbnet = BioDBNet(cache=True, show_progress=True)
    result = biodbnet.db2db(
        input_values=["4318", "1376", "2576", "10089"],
        # input_values=[str(i) for i in range(1250)],
        # annotations=["Genes"],
        input_db=Input.GENE_ID,
        output_db=Output.GENE_SYMBOL,
        taxon=Taxon.HOMO_SAPIENS,
    )
    print(result)



import json
from typing import Any, Literal, Mapping, Optional

from fast_bioservices._base import BaseModel
from fast_bioservices._fast_http import FastHTTP


class BiGG(BaseModel, FastHTTP):
    _download_url: str = "http://bigg.ucsd.edu/static/models"

    def __init__(
        self,
        max_workers: int = 4,
        cache: bool = True,
        show_progress: bool = True,
    ):
        self._url: str = "http://bigg.ucsd.edu/api/v2"
        BaseModel.__init__(self, url=self._url)
        FastHTTP.__init__(
            self,
            cache=cache,
            workers=max_workers,
            show_progress=show_progress,
            max_requests_per_second=10,
        )

    @property
    def url(self) -> str:
        return self._url

    @property
    def download_url(self) -> str:
        return self._download_url

    def version(self, temp_disable_cache: bool = False) -> Mapping[Any, Any]:
        return self._get(f"{self.url}/database_version", temp_disable_cache=temp_disable_cache)[0].json

    def models(self, temp_disable_cache: bool = False) -> Mapping[Any, Any]:
        return self._get(f"{self.url}/models", temp_disable_cache=temp_disable_cache)[0].json

    def model_details(
        self,
        model_id: str,
        temp_disable_cache: bool = False,
    ) -> Mapping[Any, Any]:
        return self._get(f"{self.url}/models/{model_id}", temp_disable_cache=temp_disable_cache)[0].json

    def json(
        self,
        model_id: str,
        temp_disable_cache: bool = False,
    ) -> Mapping[Any, Any]:
        return self._get(
            f"{self.url}/models/{model_id}/download",
            temp_disable_cache=temp_disable_cache,
        )[0].json

    def download(
        self,
        model_id: str,
        format: Literal["json", "xml", "mat", "json.gz", "xml.gz", "mat.gz"],
        download_path: Optional[str] = None,
        temp_disable_cache: bool = False,
    ) -> None:
        if download_path is None:
            download_path = f"{model_id}.{format}"
        elif not download_path.endswith(f"{model_id}.{format}"):
            download_path = f"{download_path}/{model_id}.{format}"

        response = self._get(
            f"{self.download_url}/{model_id}.{format}",
            temp_disable_cache=temp_disable_cache,
        )

        if format == "json":
            json.dump(response[0].json, open(download_path, "w"), indent=2)
        else:
            with open(download_path, "wb") as o_stream:
                o_stream.write(response[0].bytes)

    def model_reactions(
        self,
        model_id: str,
        temp_disable_cache: bool = False,
    ) -> Mapping[Any, Any]:
        return self._get(
            f"{self.url}/models/{model_id}/reactions",
            temp_disable_cache=temp_disable_cache,
        )[0].json

    def model_reaction_details(
        self,
        model_id: str,
        reaction_id: str,
        temp_disable_cache: bool = False,
    ) -> Mapping[Any, Any]:
        return self._get(
            f"{self.url}/models/{model_id}/reactions/{reaction_id}",
            temp_disable_cache=temp_disable_cache,
        )[0].json

    def model_metabolites(
        self,
        model_id: str,
        temp_disable_cache: bool = False,
    ) -> Mapping[Any, Any]:
        return self._get(
            f"{self.url}/models/{model_id}/metabolites",
            temp_disable_cache=temp_disable_cache,
        )[0].json

    def model_metabolite_details(
        self,
        model_id: str,
        metabolite_id: str,
        temp_disable_cache: bool = False,
    ) -> Mapping[Any, Any]:
        return self._get(
            f"{self.url}/models/{model_id}/metabolites/{metabolite_id}",
            temp_disable_cache=temp_disable_cache,
        )[0].json

    def model_genes(
        self,
        model_id: str,
        temp_disable_cache: bool = False,
    ) -> Mapping[Any, Any]:
        return self._get(f"{self.url}/models/{model_id}/genes", temp_disable_cache=temp_disable_cache)[0].json

    def model_gene_details(
        self,
        model_id: str,
        gene_id: str,
        temp_disable_cache: bool = False,
    ) -> Mapping[Any, Any]:
        return self._get(
            f"{self.url}/models/{model_id}/genes/{gene_id}",
            temp_disable_cache=temp_disable_cache,
        )[0].json

    def universal_reactions(
        self, temp_disable_cache: bool = False
    ) -> Mapping[Any, Any]:
        return self._get(f"{self.url}/universal/reactions", temp_disable_cache=temp_disable_cache)[0].json

    def universal_reaction_details(
        self,
        reaction_id: str,
        temp_disable_cache: bool = False,
    ) -> Mapping[Any, Any]:
        return self._get(
            f"{self.url}/universal/reactions/{reaction_id}",
            temp_disable_cache=temp_disable_cache,
        )[0].json

    def universal_metabolites(
        self, temp_disable_cache: bool = False
    ) -> Mapping[Any, Any]:
        return self._get(f"{self.url}/universal/metabolites", temp_disable_cache=temp_disable_cache)[0].json

    def universal_metabolite_details(
        self,
        metabolite_id: str,
        temp_disable_cache: bool = False,
    ) -> Mapping[Any, Any]:
        return self._get(
            f"{self.url}/universal/metabolites/{metabolite_id}",
            temp_disable_cache=temp_disable_cache,
        )[0].json

    def search(
        self,
        query: str,
        search_type: Literal["metabolites", "genes", "models", "reactions"],
        temp_disable_cache: bool = False,
    ) -> Mapping[Any, Any]:
        return self._get(
            f"{self.url}/search?query={query}&search_type={search_type}",
            temp_disable_cache=temp_disable_cache,
        )[0].json



import pandas as pd
import pytest
from fast_bioservices import BioDBNet, Input, Output, Taxon


@pytest.fixture(scope="session")
def biodbnet_instance():
    return BioDBNet(max_workers=1, cache=False, show_progress=False)


@pytest.fixture
def gene_ids(scope="session"):
    return ["4318", "1376", "2576", "10089"]


@pytest.fixture
def gene_symbols(scope="session"):
    return ["MMP9", "CPT2", "GAGE4", "KCNK7"]


def test_dbOrg(biodbnet_instance):
    result = biodbnet_instance.dbOrg(
        input_db=Input.ENSEMBL_GENE_ID,
        output_db=Output.GENE_ID,
        taxon=Taxon.HOMO_SAPIENS,
    )

    assert isinstance(result, pd.DataFrame)
    assert len(result) > 0


def test_getDirectOutputsForInput(biodbnet_instance):
    result = biodbnet_instance.getDirectOutputsForInput(Input.GENE_ID)
    print(result)

    assert len(result) > 0
    assert isinstance(result, list)


def test_getInputs(biodbnet_instance):
    result = biodbnet_instance.getInputs()
    assert len(result) > 0
    assert isinstance(result, list)


def test_getOutputsForInput(biodbnet_instance):
    result = biodbnet_instance.getOutputsForInput(Input.GENE_SYMBOL)
    assert len(result) > 0
    assert isinstance(result, list)


def test_db2db(biodbnet_instance, gene_ids, gene_symbols):
    result = biodbnet_instance.db2db(
        input_values=gene_ids,
        input_db=Input.GENE_ID,
        output_db=Output.GENE_SYMBOL,
        taxon=Taxon.HOMO_SAPIENS,
    )

    assert "Gene ID" in result.columns
    assert "Gene Symbol" in result.columns

    for id_, symbol in zip(gene_ids, gene_symbols):
        assert id_ in result["Gene ID"].values
        assert symbol in result["Gene Symbol"].values


def test_dbWalk(biodbnet_instance):
    result = biodbnet_instance.dbWalk(
        input_values=["4318", "1376", "2576", "10089"],
        db_path=[Input.GENE_ID, Input.GENE_SYMBOL],
        taxon=Taxon.HOMO_SAPIENS,
    )

    assert len(result) == 4


@pytest.mark.skip(reason="dbReport not yet implemented")
def test_dbReport(biodbnet_instance):
    biodbnet_instance.dbReport(input_values=["4318"], input_db=Input.GENE_ID, taxon=Taxon.HOMO_SAPIENS)


def test_dbFind(biodbnet_instance, gene_ids, gene_symbols):
    result = biodbnet_instance.dbFind(
        input_values=gene_ids, output_db=Output.GENE_SYNONYMS, taxon=Taxon.HOMO_SAPIENS
    )

    assert len(result) == 4
    for id_, symbol in zip(gene_ids, gene_symbols):
        assert id_ in result["InputValue"].values
        assert symbol in result["Gene Symbol"].values


def test_dbOrtho(biodbnet_instance, gene_ids):
    result = biodbnet_instance.dbOrtho(
        input_values=gene_ids,
        input_db=Input.GENE_ID,
        output_db=Output.GENE_SYMBOL,
        input_taxon=Taxon.HOMO_SAPIENS,
        output_taxon=Taxon.MUS_MUSCULUS,
    )

    assert len(result) == 4

    # symbols are from Mus Musculus, not checking those
    for id_ in zip(gene_ids):
        assert id_ in result["Gene ID"].values


@pytest.mark.skip(reason="dbAnnot tests not yet written")
def test_dbAnnot(biodbnet_instance):
    pass


@pytest.mark.skip(reason="getAllPathways tests not yet written")
def test_getAllPathways(biodbnet_instance):
    pass


@pytest.mark.skip(reason="getPathwayFromDatabase tests not yet written")
def test_getPathwayFromDatabase(biodbnet_instance):
    pass
